import re
from typing import Literal, TypedDict

from aiohttp import ClientSession

from yutto.api.types import (
    AudioUrlMeta,
    AvId,
    BvId,
    CId,
    EpisodeId,
    MediaId,
    MultiLangSubtitle,
    NoAccessError,
    SeasonId,
    UnSupportedTypeError,
    VideoUrlMeta,
)
from yutto.media.codec import VideoCodec
from yutto.utils.console.logger import Logger
from yutto.utils.fetcher import Fetcher


class BangumiListItem(TypedDict):
    id: int
    name: str
    cid: CId
    episode_id: EpisodeId
    avid: AvId
    is_section: bool  # 是否属于专区


async def get_season_id_by_media_id(session: ClientSession, media_id: MediaId) -> SeasonId:
    home_url = "https://www.bilibili.com/bangumi/media/md{media_id}".format(media_id=media_id)
    season_id = SeasonId("")
    regex_season_id = re.compile(r'"param":{"season_id":(\d+),"season_type":\d+}')
    if match_obj := regex_season_id.search(await Fetcher.fetch_text(session, home_url)):
        season_id = match_obj.group(1)
    return SeasonId(str(season_id))


async def get_season_id_by_episode_id(session: ClientSession, episode_id: EpisodeId) -> SeasonId:
    home_url = "https://www.bilibili.com/bangumi/play/ep{episode_id}".format(episode_id=episode_id)
    season_id = SeasonId
    regex_season_id = re.compile(r'"id":\d+,"ssId":(\d+)')
    if match_obj := regex_season_id.search(await Fetcher.fetch_text(session, home_url)):
        season_id = match_obj.group(1)
    return SeasonId(str(season_id))


async def get_bangumi_title(session: ClientSession, season_id: SeasonId) -> str:
    play_url = "https://www.bilibili.com/bangumi/play/ss{season_id}".format(season_id=season_id)
    regex_title = re.compile(r'<a href=".+" target="_blank" title="(.*?)" class="media-title">(?P<title>.*?)</a>')
    if match_obj := regex_title.search(await Fetcher.fetch_text(session, play_url)):
        title = match_obj.group("title")
    else:
        title = "呐，我也不知道是什么标题呢～"
    return title


async def get_bangumi_list(session: ClientSession, season_id: SeasonId) -> list[BangumiListItem]:
    list_api = "http://api.bilibili.com/pgc/view/web/season?season_id={season_id}"
    resp_json = await Fetcher.fetch_json(session, list_api.format(season_id=season_id))
    result = resp_json["result"]
    section_episodes = []
    for section in result.get("section", []):
        section_episodes += section["episodes"]
    return [
        {
            "id": i + 1,
            "name": " ".join(
                [
                    "第{}话".format(item["title"]) if re.match(r"^\d*\.?\d*$", item["title"]) else item["title"],
                    item["long_title"],
                ]
            ),
            "cid": CId(str(item["cid"])),
            "episode_id": EpisodeId(str(item["id"])),
            "avid": BvId(item["bvid"]),
            "is_section": i >= len(result["episodes"]),
        }
        for i, item in enumerate(result["episodes"] + section_episodes)
    ]


async def get_bangumi_playurl(
    session: ClientSession, avid: AvId, episode_id: EpisodeId, cid: CId
) -> tuple[list[VideoUrlMeta], list[AudioUrlMeta]]:
    play_api = "https://api.bilibili.com/pgc/player/web/playurl?avid={aid}&bvid={bvid}&ep_id={episode_id}&cid={cid}&qn=125&fnver=0&fnval=16&fourk=1"
    codecid_map: dict[Literal[7, 12], VideoCodec] = {7: "avc", 12: "hevc"}

    async with session.get(
        play_api.format(**avid.to_dict(), cid=cid, episode_id=episode_id), proxy=Fetcher.proxy
    ) as resp:
        if not resp.ok:
            raise NoAccessError("无法下载该视频（cid: {cid}）".format(cid=cid))
        resp_json = await resp.json()
        if resp_json.get("result") is None:
            raise NoAccessError("无法下载该视频（cid: {cid}），原因：{msg}".format(cid=cid, msg=resp_json.get("message")))
        if resp_json["result"].get("dash") is None:
            raise UnSupportedTypeError("该视频（cid: {cid}）尚不支持 DASH 格式".format(cid=cid))
        if resp_json["result"]["is_preview"] == 1:
            Logger.warning("视频（cid: {cid}）是预览视频".format(cid=cid))
        return (
            [
                {
                    "url": video["base_url"],
                    "mirrors": video["backup_url"] if video["backup_url"] is not None else [],
                    "codec": codecid_map[video["codecid"]],
                    "width": video["width"],
                    "height": video["height"],
                    "quality": video["id"],
                }
                for video in resp_json["result"]["dash"]["video"]
            ],
            [
                {
                    "url": audio["base_url"],
                    "mirrors": audio["backup_url"] if audio["backup_url"] is not None else [],
                    "codec": "mp4a",
                    "width": 0,
                    "height": 0,
                    "quality": audio["id"],
                }
                for audio in resp_json["result"]["dash"]["audio"]
            ],
        )


async def get_bangumi_subtitles(session: ClientSession, avid: AvId, cid: CId) -> list[MultiLangSubtitle]:
    subtitile_api = "https://api.bilibili.com/x/player/v2?cid={cid}&aid={aid}&bvid={bvid}"
    subtitile_url = subtitile_api.format(**avid.to_dict(), cid=cid)
    subtitles_info = (await Fetcher.fetch_json(session, subtitile_url))["data"]["subtitle"]
    return [
        # fmt: off
        {
            "lang": sub_info["lan_doc"],
            "lines": (await Fetcher.fetch_json(session, "https:" + sub_info["subtitle_url"]))["body"]
        }
        for sub_info in subtitles_info["subtitles"]
        # fmt: on
    ]



import re

# avid
regexp_acg_video_av = re.compile(r"https?://(www\.|m\.)?bilibili\.com/video/av(?P<aid>\d+)(\?p=(?P<page>\d+))?")

# bvid
regexp_acg_video_bv = re.compile(r"https?://(www\.|m\.)?bilibili\.com/video/(?P<bvid>(bv|BV)\w+)(\?p=(?P<page>\d+))?")

# media id
regexp_bangumi_md = re.compile(r"https?://(www\.|m\.)?bilibili\.com/bangumi/media/md(?P<media_id>\d+)")

# episode id
regexp_bangumi_ep = re.compile(r"https?://(www\.|m\.)?bilibili\.com/bangumi/play/ep(?P<episode_id>\d+)")

# season id
regexp_bangumi_ss = re.compile(r"https?://(www\.|m\.)?bilibili\.com/bangumi/play/ss(?P<season_id>\d+)")



from enum import Enum
from typing import Any, Literal


class Media(Enum):
    VIDEO = 0
    AUDIO = 30200


VideoQuality = Literal[125, 120, 116, 112, 80, 74, 64, 32, 16]
AudioQuality = Literal[30280, 30232, 30216]

video_quality_priority_default: list[VideoQuality] = [125, 120, 116, 112, 80, 74, 64, 32, 16]
audio_quality_priority_default: list[AudioQuality] = [30280, 30232, 30216]

video_quality_map = {
    125: {
        "description": "HDR 真彩",
        "width": 3840,
        "height": 1920,
    },
    120: {
        "description": "4K 超清",
        "width": 3840,
        "height": 1920,
    },
    116: {
        "description": "1080P 60帧",
        "width": 2160,
        "height": 1080,
    },
    112: {
        "description": "1080P 高码率",
        "width": 2160,
        "height": 1080,
    },
    80: {
        "description": "1080P 高清",
        "width": 2160,
        "height": 1080,
    },
    74: {
        "description": "720P 60帧",
        "width": 1440,
        "height": 720,
    },
    64: {
        "description": "720P 高清",
        "width": 1440,
        "height": 720,
    },
    32: {
        "description": "480P 清晰",
        "width": 960,
        "height": 480,
    },
    16: {
        "description": "360P 流畅",
        "width": 720,
        "height": 360,
    },
}

audio_quality_map = {
    30280: {
        "description": "320kbps",
        "bitrate": 320,
    },
    30232: {
        "description": "128kbps",
        "bitrate": 128,
    },
    30216: {
        "description": "64kbps",
        "bitrate": 64,
    },
    0: {"description": "Unknown", "bitrate": 0},
}


def gen_quality_priority(quality: Any, quality_priority: list[Any]) -> list[Any]:
    """ 根据默认先降后升的清晰度机制生成清晰度序列 """

    return quality_priority[quality_priority.index(quality) :] + list(
        reversed(quality_priority[: quality_priority.index(quality)])
    )


def gen_audio_quality_priority(quality: AudioQuality) -> list[AudioQuality]:
    return gen_quality_priority(quality, audio_quality_priority_default)


def gen_video_quality_priority(quality: VideoQuality) -> list[VideoQuality]:
    return gen_quality_priority(quality, video_quality_priority_default)



from typing import NamedTuple, TypedDict

from yutto.media.codec import AudioCodec, VideoCodec
from yutto.media.quality import AudioQuality, VideoQuality
from yutto.utils.subtitle import SubtitleData


class BilibiliId(NamedTuple):
    value: str

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other: "BilibiliId") -> bool:
        return self.value == other.value


class AvId(BilibiliId):
    def to_dict(self) -> dict[str, str]:
        raise NotImplementedError("请不要直接使用 AvId")


class AId(AvId):
    def to_dict(self):
        return {"aid": self.value, "bvid": ""}


class BvId(AvId):
    def to_dict(self):
        return {
            "aid": "",
            "bvid": self.value,
        }


class CId(BilibiliId):
    def to_dict(self):
        return {"cid": self.value}


class EpisodeId(BilibiliId):
    def to_dict(self):
        return {"episode_id": self.value}


class MediaId(BilibiliId):
    def to_dict(self):
        return {"media_id": self.value}


class SeasonId(BilibiliId):
    def to_dict(self):
        return {"season_id": self.value}


class HttpStatusError(Exception):
    pass


class NoAccessError(Exception):
    pass


class UnSupportedTypeError(Exception):
    pass


class VideoUrlMeta(TypedDict):
    url: str
    mirrors: list[str]
    codec: VideoCodec
    width: int
    height: int
    quality: VideoQuality


class AudioUrlMeta(TypedDict):
    url: str
    mirrors: list[str]
    codec: AudioCodec
    width: int
    height: int
    quality: AudioQuality


class MultiLangSubtitle(TypedDict):
    lang: str
    lines: SubtitleData


if __name__ == "__main__":
    aid = AId("add")
    cid = CId("xxx")
    print("?aid={aid}&bvid={bvid}&cid={cid}".format(**aid.to_dict(), **cid.to_dict()))



import os
import re
from html import unescape
from typing import Union, Literal

path_template_variables = ["title", "id", "name"]
PathTemplateVariable = Literal["title", "id", "name"]

_count: int = 0


def repair_filename(filename: str) -> str:
    """ 修复不合法的文件名 """

    def to_full_width_chr(matchobj: re.Match[str]) -> str:
        char = matchobj.group(0)
        full_width_char = chr(ord(char) + ord("？") - ord("?"))
        return full_width_char

    # 路径非法字符，转全角
    regex_path = re.compile(r'[\\/:*?"<>|]')
    # 空格类字符，转空格
    regex_spaces = re.compile(r"\s+")
    # 不可打印字符，移除
    regex_non_printable = re.compile(
        r"[\001\002\003\004\005\006\007\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f"
        r"\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a]"
    )

    # 由于部分内容可能是从 HTML 解析的，所以使用 html 反转义
    filename = unescape(filename)
    filename = regex_path.sub(to_full_width_chr, filename)
    filename = regex_spaces.sub(" ", filename)
    filename = regex_non_printable.sub("", filename)
    filename = filename.strip()
    if not filename:
        filename = "未命名文件_{:04}".format(_count)
        _count += 1
    return filename


def resolve_path_template(
    path_template: str, auto_path_template: str, template_dict: dict[PathTemplateVariable, Union[int, str]]
) -> str:
    # 保证所有传进来的值都满足路径要求
    for key, value in template_dict.items():
        # 只对字符串值修改，int 型不修改以适配高级模板
        if isinstance(value, str):
            template_dict[key] = repair_filename(value)
    return path_template.format(auto=auto_path_template.format(**template_dict), **template_dict)



import argparse
import os
import sys

import aiohttp

from yutto.api.acg_video import get_acg_video_list, get_acg_video_playurl, get_acg_video_subtitles, get_acg_video_title
from yutto.api.bangumi import (
    get_bangumi_list,
    get_bangumi_playurl,
    get_bangumi_subtitles,
    get_bangumi_title,
    get_season_id_by_episode_id,
    get_season_id_by_media_id,
)
from yutto.api.danmaku import get_danmaku
from yutto.api.types import AId, AudioUrlMeta, BvId, EpisodeId, MediaId, MultiLangSubtitle, SeasonId, VideoUrlMeta
from yutto.cli import check_options
from yutto.processor.downloader import download_video
from yutto.processor.filter import parse_episodes
from yutto.processor.path_resolver import resolve_path_template
from yutto.processor.urlparser import (
    regexp_acg_video_av,
    regexp_acg_video_bv,
    regexp_bangumi_ep,
    regexp_bangumi_md,
    regexp_bangumi_ss,
)
from yutto.utils.console.logger import Badge, Logger
from yutto.utils.danmaku import DanmakuData, EmptyDanmakuData
from yutto.utils.fetcher import Fetcher
from yutto.utils.functiontools.sync import sync


def add_arguments(parser: argparse.ArgumentParser):
    parser.add_argument("url", help="视频主页 url")
    parser.add_argument("-p", "--episodes", default="^~$", help="选集")
    parser.add_argument("-s", "--with-section", action="store_true", help="同时下载附加剧集（PV、预告以及特别篇等专区内容）")
    parser.set_defaults(action=run)


@sync
async def run(args: argparse.Namespace):
    check_options.check_batch_options(args)
    async with aiohttp.ClientSession(
        headers=Fetcher.headers,
        cookies=Fetcher.cookies,
        trust_env=Fetcher.trust_env,
        timeout=aiohttp.ClientTimeout(total=5),
    ) as session:
        url: str = args.url
        url = await Fetcher.get_redirected_url(session, url)
        download_list: list[
            tuple[list[VideoUrlMeta], list[AudioUrlMeta], str, str, list[MultiLangSubtitle], DanmakuData]
        ] = []
        if (
            (match_obj := regexp_bangumi_ep.match(url))
            or (match_obj := regexp_bangumi_ss.match(url))
            or (match_obj := regexp_bangumi_md.match(url))
        ):
            # 匹配为番剧
            if "episode_id" in match_obj.groupdict().keys():
                episode_id = EpisodeId(match_obj.group("episode_id"))
                season_id = await get_season_id_by_episode_id(session, episode_id)
            elif "season_id" in match_obj.groupdict().keys():
                season_id = SeasonId(match_obj.group("season_id"))
            else:
                media_id = MediaId(match_obj.group("media_id"))
                season_id = await get_season_id_by_media_id(session, media_id)
            title = await get_bangumi_title(session, season_id)
            Logger.custom(title, Badge("番剧", fore="black", back="cyan"))
            bangumi_list = await get_bangumi_list(session, season_id)
            # 如果没有 with_section 则不需要专区内容
            bangumi_list = list(filter(lambda item: args.with_section or not item["is_section"], bangumi_list))
            # 选集过滤
            episodes = parse_episodes(args.episodes, len(bangumi_list))
            bangumi_list = list(filter(lambda item: item["id"] in episodes, bangumi_list))
            for i, bangumi_item in enumerate(bangumi_list):
                Logger.status.set("正在努力解析第 {}/{} 个视频".format(i + 1, len(bangumi_list)))
                avid = bangumi_item["avid"]
                cid = bangumi_item["cid"]
                episode_id = bangumi_item["episode_id"]
                name = bangumi_item["name"]
                id = bangumi_item["id"]
                videos, audios = await get_bangumi_playurl(session, avid, episode_id, cid)
                # fmt: off
                subpath = resolve_path_template(
                    args.subpath_template,
                    "{title}/{name}",
                    {
                        "title": title,
                        "id": id,
                        "name": name
                    })
                # fmt: on
                output_dir, filename = os.path.split(os.path.join(args.dir, subpath))
                subtitles = await get_bangumi_subtitles(session, avid, cid) if not args.no_subtitle else []
                danmaku = (
                    await get_danmaku(session, cid, args.danmaku_format) if not args.no_danmaku else EmptyDanmakuData
                )
                download_list.append((videos, audios, output_dir, filename, subtitles, danmaku))
        elif (match_obj := regexp_acg_video_av.match(url)) or (match_obj := regexp_acg_video_bv.match(url)):
            # 匹配为投稿视频
            if "aid" in match_obj.groupdict().keys():
                avid = AId(match_obj.group("aid"))
            else:
                avid = BvId(match_obj.group("bvid"))
            title = await get_acg_video_title(session, avid)
            Logger.custom(title, Badge("投稿视频", fore="black", back="cyan"))
            acg_video_list = await get_acg_video_list(session, avid)
            # 选集过滤
            episodes = parse_episodes(args.episodes, len(acg_video_list))
            acg_video_list = list(filter(lambda item: item["id"] in episodes, acg_video_list))
            for i, acg_video_item in enumerate(acg_video_list):
                Logger.status.set("正在努力解析第 {}/{} 个视频".format(i + 1, len(acg_video_list)))
                cid = acg_video_item["cid"]
                name = acg_video_item["name"]
                id = acg_video_item["id"]
                videos, audios = await get_acg_video_playurl(session, avid, cid)
                # fmt: off
                subpath = resolve_path_template(
                    args.subpath_template,
                    "{title}/{name}",
                    {
                        "title": title,
                        "id": id,
                        "name": name
                    })
                # fmt: on
                output_dir, filename = os.path.split(os.path.join(args.dir, subpath))
                subtitles = await get_acg_video_subtitles(session, avid, cid) if not args.no_subtitle else []
                danmaku = (
                    await get_danmaku(session, cid, args.danmaku_format) if not args.no_danmaku else EmptyDanmakuData
                )
                download_list.append((videos, audios, output_dir, filename, subtitles, danmaku))
        else:
            Logger.error("url 不正确～")
            sys.exit(1)
        for videos, audios, output_dir, filename, subtitles, danmaku in download_list:

            await download_video(
                session,
                videos,
                audios,
                output_dir,
                filename,
                subtitles,
                danmaku,
                {
                    "require_video": args.require_video,
                    "video_quality": args.video_quality,
                    "video_download_codec": args.vcodec.split(":")[0],
                    "video_save_codec": args.vcodec.split(":")[1],
                    "require_audio": args.require_audio,
                    "audio_quality": args.audio_quality,
                    "audio_download_codec": args.acodec.split(":")[0],
                    "audio_save_codec": args.acodec.split(":")[1],
                    "overwrite": args.overwrite,
                    "block_size": int(args.block_size * 1024 * 1024),
                    "num_workers": args.num_workers,
                },
            )



import heapq
import os
from dataclasses import dataclass, field
from typing import Optional

import aiofiles

from yutto.utils.console.logger import Logger


@dataclass(order=True)
class BufferChunk:
    offset: int
    data: bytes = field(compare=False)


class AsyncFileBuffer:
    def __init__(self):
        self.file_path = ""
        self.file_obj: Optional[aiofiles.threadpool.binary.AsyncBufferedIOBase] = None
        self.buffer = list[BufferChunk]()
        self.written_size = 0

    @classmethod
    async def create(cls, file_path: str, overwrite: bool = False):
        self = cls()
        self.file_path = file_path
        if overwrite and os.path.exists(file_path):
            os.remove(file_path)
        self.written_size = os.path.getsize(file_path) if not overwrite and os.path.exists(file_path) else 0
        self.file_obj = await aiofiles.open(file_path, "ab")
        return self

    async def write(self, chunk: bytes, offset: int):
        buffer_chunk = BufferChunk(offset, chunk)
        # 使用堆结构，保证第一个元素始终最小
        heapq.heappush(self.buffer, buffer_chunk)
        while self.buffer and self.buffer[0].offset <= self.written_size:
            assert self.file_obj is not None
            ready_to_write_chunk = heapq.heappop(self.buffer)
            if ready_to_write_chunk.offset < self.written_size:
                Logger.error("交叠的块范围 {} < {}，舍弃！".format(ready_to_write_chunk.offset, self.written_size))
                continue
            await self.file_obj.write(ready_to_write_chunk.data)
            self.written_size += len(ready_to_write_chunk.data)

    async def close(self):
        if self.buffer:
            Logger.error("buffer 尚未清空")
        if self.file_obj is not None:
            await self.file_obj.close()
        else:
            Logger.error("未预期的结果：未曾创建文件对象")



import asyncio
import platform
from typing import Any, Coroutine, Iterable

from yutto.utils.console.logger import Logger

CoroutineTask = Coroutine[Any, Any, Any]


def initial_async_policy():
    if platform.system() == "Windows":
        Logger.debug("Windows 平台，单独设置 EventLoopPolicy")
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


def install_uvloop():
    try:
        import uvloop
    except ImportError:
        Logger.warning("未安装 uvloop，无法使用其加速协程")
    else:
        uvloop.install()
        Logger.info("成功使用 uvloop 加速协程")


def parallel(funcs: Iterable[CoroutineTask]):
    return [asyncio.create_task(func) for func in funcs]


def parallel_with_limit(funcs: Iterable[CoroutineTask], num_workers: int = 4):
    tasks = asyncio.Queue[CoroutineTask]()
    for func in funcs:
        tasks.put_nowait(func)

    async def worker():
        while True:
            if not tasks.empty():
                task = await tasks.get()
                await task
            else:
                break

    return [asyncio.create_task(worker()) for _ in range(num_workers)]



from typing import TypedDict

from aiohttp import ClientSession

from yutto.api.types import AId, AvId, BvId, CId, EpisodeId
from yutto.processor.urlparser import regexp_bangumi_ep
from yutto.utils.fetcher import Fetcher


class VideoInfo(TypedDict):
    avid: AvId
    aid: AId
    bvid: BvId
    episode_id: EpisodeId
    is_bangumi: bool
    cid: CId
    picture: str
    title: str


async def get_video_info(session: ClientSession, avid: AvId) -> VideoInfo:
    info_api = "http://api.bilibili.com/x/web-interface/view?aid={aid}&bvid={bvid}"
    res_json = await Fetcher.fetch_json(session, info_api.format(**avid.to_dict()))
    res_json_data = res_json.get("data")
    assert res_json_data is not None, "响应数据无 data 域"
    episode_id = EpisodeId("")
    if res_json_data.get("redirect_url") and (ep_match := regexp_bangumi_ep.match(res_json_data["redirect_url"])):
        episode_id = EpisodeId(ep_match.group("episode_id"))
    return {
        "avid": BvId(res_json_data["bvid"]),
        "aid": AId(str(res_json_data["aid"])),
        "bvid": BvId(res_json_data["bvid"]),
        "episode_id": episode_id,
        "is_bangumi": bool(episode_id),
        "cid": CId(str(res_json_data["cid"])),
        "picture": res_json_data["pic"],
        "title": res_json_data["title"],
    }


async def is_vip(session: ClientSession) -> bool:
    info_api = "https://api.bilibili.com/x/web-interface/nav"
    res_json = await Fetcher.fetch_json(session, info_api)
    res_json_data = res_json.get("data")
    if res_json_data.get("vipStatus") == 1:
        return True
    return False



import argparse

from yutto.media.quality import audio_quality_priority_default, video_quality_priority_default


def add_basic_arguments(parser: argparse.ArgumentParser):
    parser.add_argument("-n", "--num-workers", type=int, default=8, help="同时用于下载的最大 Worker 数")
    parser.add_argument(
        "-q",
        "--video-quality",
        default=125,
        choices=video_quality_priority_default,
        type=int,
        help="视频清晰度等级（125:HDR, 120:4K, 116:1080P60, 112:1080P+, 80:1080P, 74:720P60, 64:720P, 32:480P, 16:360P）",
    )
    parser.add_argument(
        "-aq",
        "--audio-quality",
        default=30280,
        choices=audio_quality_priority_default,
        type=int,
        help="音频码率等级（30280:320kbps, 30232:128kbps, 30216:64kbps）",
    )
    parser.add_argument(
        "--vcodec", default="avc:copy", metavar="DOWNLOAD_VCODEC:SAVE_VCODEC", help="视频编码格式（<下载格式>:<生成格式>）"
    )
    parser.add_argument(
        "--acodec", default="mp4a:copy", metavar="DOWNLOAD_ACODEC:SAVE_ACODEC", help="音频编码格式（<下载格式>:<生成格式>）"
    )
    parser.add_argument("--only-video", dest="require_audio", action="store_false", help="只下载视频")
    parser.add_argument("--only-audio", dest="require_video", action="store_false", help="只下载音频")
    parser.add_argument("-df", "--danmaku-format", default="ass", choices=["xml", "ass", "protobuf"], help="弹幕类型")
    parser.add_argument("-b", "--block-size", default=0.5, type=float, help="分块下载时各块大小，单位为 MiB，默认为 0.5MiB")
    parser.add_argument("-w", "--overwrite", action="store_true", help="强制覆盖已下载内容")
    parser.add_argument("-x", "--proxy", default="auto", help="设置代理（auto 为系统代理、no 为不使用代理、当然也可以设置代理值）")
    parser.add_argument("-d", "--dir", default="./", help="下载目录，默认为运行目录")
    parser.add_argument("-c", "--sessdata", default="", help="Cookies 中的 SESSDATA 字段")
    parser.add_argument("-tp", "--subpath-template", default="{auto}", help="多级目录的存储路径模板")
    parser.add_argument("--no-danmaku", action="store_true", help="不生成弹幕文件")
    parser.add_argument("--no-subtitle", action="store_true", help="不生成字幕文件")
    parser.add_argument("--embed-danmaku", action="store_true", help="（待实现）将弹幕文件嵌入到视频中")
    parser.add_argument("--embed-subtitle", default=None, help="（待实现）将字幕文件嵌入到视频中（需输入语言代码）")
    parser.add_argument("--no-color", action="store_true", help="不使用颜色")
    parser.add_argument("--debug", action="store_true", help="启用 debug 模式")



import argparse
import os
import sys

import aiohttp

from yutto.api.acg_video import get_acg_video_list, get_acg_video_playurl, get_acg_video_subtitles, get_acg_video_title
from yutto.api.bangumi import (
    get_bangumi_list,
    get_bangumi_playurl,
    get_bangumi_subtitles,
    get_bangumi_title,
    get_season_id_by_episode_id,
)
from yutto.api.danmaku import get_danmaku
from yutto.api.types import AId, BvId, EpisodeId
from yutto.processor.downloader import download_video
from yutto.processor.path_resolver import resolve_path_template
from yutto.processor.urlparser import (
    regexp_acg_video_av,
    regexp_acg_video_bv,
    regexp_bangumi_ep,
)
from yutto.utils.console.logger import Badge, Logger
from yutto.utils.danmaku import EmptyDanmakuData
from yutto.utils.fetcher import Fetcher
from yutto.utils.functiontools.sync import sync


def add_arguments(parser: argparse.ArgumentParser):
    parser.add_argument("url", help="视频主页 url")
    parser.set_defaults(action=run)


@sync
async def run(args: argparse.Namespace):
    async with aiohttp.ClientSession(
        headers=Fetcher.headers,
        cookies=Fetcher.cookies,
        trust_env=Fetcher.trust_env,
        timeout=aiohttp.ClientTimeout(total=5),
    ) as session:
        url: str = args.url
        url = await Fetcher.get_redirected_url(session, url)
        if match_obj := regexp_bangumi_ep.match(url):
            # 匹配为番剧
            episode_id = EpisodeId(match_obj.group("episode_id"))
            season_id = await get_season_id_by_episode_id(session, episode_id)
            title = await get_bangumi_title(session, season_id)
            Logger.custom(title, Badge("番剧", fore="black", back="cyan"))
            bangumi_list = await get_bangumi_list(session, season_id)
            for bangumi_item in bangumi_list:
                if bangumi_item["episode_id"] == episode_id:
                    avid = bangumi_item["avid"]
                    cid = bangumi_item["cid"]
                    name = bangumi_item["name"]
                    id = bangumi_item["id"]
                    break
            else:
                Logger.error("在列表中未找到该剧集")
                sys.exit(1)
            auto_subpath_template = "{name}"
            videos, audios = await get_bangumi_playurl(session, avid, episode_id, cid)
            subtitles = await get_bangumi_subtitles(session, avid, cid) if not args.no_subtitle else []
            danmaku = await get_danmaku(session, cid, args.danmaku_format) if not args.no_danmaku else EmptyDanmakuData
        elif (match_obj := regexp_acg_video_av.match(url)) or (match_obj := regexp_acg_video_bv.match(url)):
            # 匹配为投稿视频
            page: int = 1
            if "aid" in match_obj.groupdict().keys():
                avid = AId(match_obj.group("aid"))
            else:
                avid = BvId(match_obj.group("bvid"))
            if match_obj.group("page") is not None:
                page = int(match_obj.group("page"))
            title = await get_acg_video_title(session, avid)
            Logger.custom(title, Badge("投稿视频", fore="black", back="cyan"))
            acg_video_list = await get_acg_video_list(session, avid)
            cid = acg_video_list[page - 1]["cid"]
            name = acg_video_list[page - 1]["name"]
            id = acg_video_list[page - 1]["id"]
            auto_subpath_template = "{title}"
            videos, audios = await get_acg_video_playurl(session, avid, cid)
            subtitles = await get_acg_video_subtitles(session, avid, cid) if not args.no_subtitle else []
            danmaku = await get_danmaku(session, cid, args.danmaku_format) if not args.no_danmaku else EmptyDanmakuData
        else:
            Logger.error("url 不正确～")
            sys.exit(1)
        # fmt: off
        subpath = resolve_path_template(
            args.subpath_template,
            auto_subpath_template,
            {
                "title": title,
                "id": id,
                "name": name
            })
        # fmt: on
        output_dir, filename = os.path.split(os.path.join(args.dir, subpath))

        await download_video(
            session,
            videos,
            audios,
            output_dir,
            filename,
            subtitles,
            danmaku,
            {
                "require_video": args.require_video,
                "video_quality": args.video_quality,
                "video_download_codec": args.vcodec.split(":")[0],
                "video_save_codec": args.vcodec.split(":")[1],
                "require_audio": args.require_audio,
                "audio_quality": args.audio_quality,
                "audio_download_codec": args.acodec.split(":")[0],
                "audio_save_codec": args.acodec.split(":")[1],
                "overwrite": args.overwrite,
                "block_size": int(args.block_size * 1024 * 1024),
                "num_workers": args.num_workers,
            },
        )



import asyncio
import math
import time
from typing import Union, Optional

from yutto.utils.console.colorful import colored_string, Color, RGBColor
from yutto.utils.console.formatter import size_format
from yutto.utils.console.logger import Logger
from yutto.utils.file_buffer import AsyncFileBuffer


class ProgressBar:
    def __init__(self, symbols: Union[str, list[str]] = "▏▎▍▌▋▊▉█", remaining_symbol: str = " ", width: int = 50):
        super().__init__()
        self.width = width
        self.symbols = symbols
        self.remaining_symbol = remaining_symbol
        assert len(symbols) >= 2, "symbols 至少为 2 个"
        self.num_symbol = len(symbols)

    def render(
        self,
        data: float,
        bar_fore_color: Optional[Color] = None,
        bar_back_color: Optional[Color] = None,
        remaining_bar_fore_color: Optional[Color] = None,
        remaining_bar_back_color: Optional[Color] = None,
    ) -> str:
        if data == 1:
            return self.symbols[-1] * self.width
        length: float = self.width * data
        length_int: int = int(length)
        length_float: float = length - length_int

        return colored_string(
            length_int * self.symbols[-1] + self.symbols[math.floor(length_float * self.num_symbol)],
            fore=bar_fore_color,
            back=bar_back_color,
        ) + colored_string(
            (self.width - length_int - 1) * self.remaining_symbol,
            fore=remaining_bar_fore_color,
            back=remaining_bar_back_color,
        )


async def show_progress(file_buffers: list[AsyncFileBuffer], total_size: int):
    file_buffers = list(filter(lambda x: x is not None, file_buffers))
    t = time.time()
    size = sum([file_buffer.written_size for file_buffer in file_buffers])
    progress_bar = ProgressBar("╸━", "━")
    while True:
        size_in_buffer: int = sum(
            [sum([len(chunk.data) for chunk in file_buffer.buffer]) for file_buffer in file_buffers]
        )
        num_blocks_in_buffer: int = sum([len(file_buffer.buffer) for file_buffer in file_buffers])
        size_written: int = sum([file_buffer.written_size for file_buffer in file_buffers])

        t_now = time.time()
        size_now = size_written + size_in_buffer
        speed = (size_now - size) / (t_now - t + 10 ** -6)

        # 默认颜色为青色
        # 当速度过快导致 buffer 中的块数过多时（>1000 块），使用红色进行警告
        # 在速度高于 8MiB/s 时，使用绿色示意高速下载中
        color = "red" if num_blocks_in_buffer > 1000 else ("green" if speed >= 8 * 1024 * 1024 else "cyan")
        bar = progress_bar.render(
            size_now / total_size, bar_fore_color=color, remaining_bar_fore_color=RGBColor(64, 64, 64)
        )
        # fmt: off
        Logger.status.set(
            "{} {:>10}/{:>10} {:>10}/s  ".format(
                bar,
                size_format(size_now),
                size_format(total_size),
                size_format(speed),
            )
        )

        t, size = t_now, size_now
        await asyncio.sleep(0.5)
        if total_size == size:
            break



import os
import sys
from shutil import rmtree

from yutto.__version__ import __version__
from setuptools import setup, find_packages, Command

here = os.path.abspath(os.path.dirname(__file__))


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s: str):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(__version__))
        os.system("git push --tags")

        sys.exit()


def get_long_description():
    with open("README.md", "r", encoding="utf-8") as f:
        desc = f.read()
    return desc


setup(
    name="yutto",
    version=__version__,
    description="yutto 一个可爱且任性的 B 站视频下载器",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    keywords="python bilibili video download spider danmaku",
    author="Nyakku Shigure",
    author_email="sigure.qaq@gmail.com",
    url="https://github.com/SigureMo/yutto",
    license="GPLv3",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=True,
    python_requires=">=3.9.0",
    setup_requires=["wheel"],
    install_requires=["aiohttp", "aiofiles", "asyncio", "biliass"],
    entry_points={"console_scripts": ["yutto = yutto.__main__:main"]},
    cmdclass={
        "upload": UploadCommand,
    },
)



import argparse
import sys

from yutto.__version__ import __version__
from yutto.cli import basic, batch_get, check_options, get, info
from yutto.utils.console.logger import Badge, Logger
from yutto.utils.ffmpeg import FFmpeg


def main():
    parser = argparse.ArgumentParser(description="yutto 一个可爱且任性的 B 站视频下载器", prog="yutto")
    parser.add_argument("-v", "--version", action="version", version="%(prog)s {}".format(__version__))

    subparsers = parser.add_subparsers()
    # 子命令 get
    parser_get = subparsers.add_parser("get", help="获取单个视频")
    basic.add_basic_arguments(parser_get)
    get.add_arguments(parser_get)
    # 子命令 info
    # TODO
    # 子命令 batch
    parser_batch = subparsers.add_parser("batch", help="批量获取视频（需使用其子命令 get/info）")
    subparsers_batch = parser_batch.add_subparsers()
    # 子命令 batch get
    parser_batch_get = subparsers_batch.add_parser("get", help="批量获取视频")
    basic.add_basic_arguments(parser_batch_get)
    batch_get.add_arguments(parser_batch_get)
    # 子命令 batch info
    # TODO
    # 执行各自的 action
    args = parser.parse_args()
    if "action" in args:
        check_options.check_basic_options(args)
        args.action(args)
    else:
        Logger.error("未指定子命令 (get, info, batch)")
        python_version = "{}.{}.{}".format(sys.version_info[0], sys.version_info[1], sys.version_info[2])
        ffmpeg_version = FFmpeg().version
        yutto_version = __version__
        Logger.print(
            Badge("Python", fore="black", back="green"), Badge(python_version, fore="green", back="black"), sep=""
        )
        Logger.print(Badge("yutto", fore="black", back="blue"), Badge(yutto_version, fore="blue", back="black"), sep="")
        Logger.print(
            Badge("ffmpeg", fore="black", back="cyan"), Badge(ffmpeg_version, fore="cyan", back="black"), sep=""
        )
        parser.print_help()


if __name__ == "__main__":
    main()
